# Generated by Django 5.2.5 on 2025-08-21 11:21
import taggit.managers
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("extras", "0128_tableconfig"),
        ("ipfabric_netbox", "0015_ipfabricingestionissue"),
    ]

    operations = [
        migrations.AddField(
            model_name="ipfabricsnapshot",
            name="tags",
            field=taggit.managers.TaggableManager(
                through="extras.TaggedItem", to="extras.Tag"
            ),
        ),
        migrations.AlterField(
            model_name="ipfabricsnapshot",
            name="created",
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
        migrations.AlterField(
            model_name="ipfabricsnapshot",
            name="last_updated",
            field=models.DateTimeField(auto_now=True, null=True),
        ),
    ]
